@@ -40,7 +40,6 @@
  * text to make the parser recognize only the rules corresponding to CLI commands.
  */
 
-#include <setjmp.h>
 #include <stdarg.h>
 
 #undef LOCAL_DEBUG
@@ -49,6 +48,7 @@
 #include "nest/route.h"
 #include "nest/protocol.h"
 #include "nest/iface.h"
+#include "nest/mpls.h"
 #include "lib/resource.h"
 #include "lib/string.h"
 #include "lib/event.h"
@@ -60,17 +60,18 @@
 
 static jmp_buf conf_jmpbuf;
 
-struct config *config, *new_config;
+config_ref config;
+_Thread_local struct config *new_config;
 pool *config_pool;
 
-static struct config *old_config;	/* Old configuration */
-static struct config *future_config;	/* New config held here if recon requested during recon */
+struct config *old_config;		/* Old configuration */
+static config_ref future_config;	/* New config held here if recon requested during recon */
 static int old_cftype;			/* Type of transition old_config -> config (RECONFIG_SOFT/HARD) */
 static int future_cftype;		/* Type of scheduled transition, may also be RECONFIG_UNDO */
 /* Note that when future_cftype is RECONFIG_UNDO, then future_config is NULL,
    therefore proper check for future scheduled config checks future_cftype */
 
-static event *config_event;		/* Event for finalizing reconfiguration */
+static void config_done(void);
 static timer *config_timer;		/* Timer for scheduled configuration rollback */
 
 /* These are public just for cmd_show_status(), should not be accessed elsewhere */
@@ -79,6 +80,14 @@ int configuring;			/* Reconfiguration is running */
 int undo_available;			/* Undo was not requested from last reconfiguration */
 /* Note that both shutting_down and undo_available are related to requests, not processing */
 
+static void
+config_obstacles_cleared(struct callback *_ UNUSED)
+{
+  ASSERT_DIE(birdloop_inside(&main_birdloop));
+  ASSERT_DIE(configuring);
+  config_done();
+}
+
 /**
  * config_alloc - allocate a new configuration
  * @name: name of the config
@@ -90,7 +99,7 @@ int undo_available;			/* Undo was not requested from last reconfiguration */
 struct config *
 config_alloc(const char *name)
 {
-  pool *p = rp_new(config_pool, "Config");
+  pool *p = rp_new(config_pool, the_bird_domain.the_bird, "Config");
   linpool *l = lp_new_default(p);
   struct config *c = lp_allocz(l, sizeof(struct config));
 
@@ -101,15 +110,16 @@ config_alloc(const char *name)
 
   init_list(&c->tests);
   init_list(&c->symbols);
-  c->mrtdump_file = -1; /* Hack, this should be sysdep-specific */
   c->pool = p;
   c->mem = l;
   c->file_name = ndup;
-  c->load_time = current_time();
   c->tf_route = c->tf_proto = TM_ISO_SHORT_MS;
   c->tf_base = c->tf_log = TM_ISO_LONG_MS;
   c->gr_wait = DEFAULT_GR_WAIT;
 
+  callback_init(&c->obstacles_cleared, config_obstacles_cleared, &main_birdloop);
+  obstacle_target_init(&c->obstacles, &c->obstacles_cleared, p, "Config");
+
   return c;
 }
 
@@ -136,9 +146,11 @@ config_parse(struct config *c)
   if (setjmp(conf_jmpbuf))
     goto cleanup;
 
-  cf_lex_init(0, c);
+  cf_lex_init(NULL, c);
+  filter_preconfig(c);
   sysdep_preconfig(c);
   protos_preconfig(c);
+  mpls_preconfig(c);
   rt_preconfig(c);
   cf_parse();
   rt_postconfig(c);
@@ -146,11 +158,6 @@ config_parse(struct config *c)
   if (EMPTY_LIST(c->protos))
     cf_error("No protocol is specified in the config file");
 
-  /*
-  if (!c->router_id)
-    cf_error("Router ID must be configured manually");
-  */
-
   done = 1;
 
 cleanup:
@@ -167,21 +174,19 @@ cleanup:
  * it parses a CLI command. See the CLI module for more information.
  */
 int
-cli_parse(struct config *c)
+cli_parse(struct config *main_config, struct config *c)
 {
   int done = 0;
-  c->fallback = config;
   new_config = c;
   cfg_mem = c->mem;
   if (setjmp(conf_jmpbuf))
     goto cleanup;
 
-  cf_lex_init(1, c);
+  cf_lex_init(main_config, c);
   cf_parse();
   done = 1;
 
 cleanup:
-  c->fallback = NULL;
   new_config = NULL;
   cfg_mem = NULL;
   return done;
@@ -200,9 +205,10 @@ config_free(struct config *c)
   if (!c)
     return;
 
-  ASSERT(!c->obstacle_count);
+  synchronize_rcu();
+  ASSERT_DIE(!obstacle_target_count(&c->obstacles));
 
-  rfree(c->pool);
+  rp_free(c->pool);
 }
 
 /**
@@ -216,124 +222,163 @@ config_free(struct config *c)
 void
 config_free_old(void)
 {
-  if (!old_config || old_config->obstacle_count)
-    return;
-
   tm_stop(config_timer);
   undo_available = 0;
 
-  config_free(old_config);
-  old_config = NULL;
+  if (!configuring)
+  {
+    config_free(old_config);
+    old_config = NULL;
+  }
 }
 
-void
-config_add_obstacle(struct config *c)
-{
-  DBG("+++ adding obstacle %d\n", c->obstacle_count);
-  c->obstacle_count++;
-}
+struct global_runtime global_runtime_internal[2] = {{
+  .tf_log = {
+    .fmt1 = "%F %T.%3f",
+  },
+}};
+struct global_runtime * _Atomic global_runtime = &global_runtime_internal[0];
 
-void
-config_del_obstacle(struct config *c)
-{
-  DBG("+++ deleting obstacle %d\n", c->obstacle_count);
-  c->obstacle_count--;
-  if (!c->obstacle_count && (c != config))
-    ev_schedule(config_event);
-}
-
-static int
+static void
 global_commit(struct config *new, struct config *old)
 {
-  if (!new->hostname)
+  /* Updating the global runtime. */
+  struct global_runtime *og = atomic_load_explicit(&global_runtime, memory_order_relaxed);
+  struct global_runtime *ng = &global_runtime_internal[og == &global_runtime_internal[0]];
+  ASSERT_DIE(ng != og);
+
+#define COPY(x)	ng->x = new->x;
+  MACRO_FOREACH(COPY,
+      tf_route,
+      tf_proto,
+      tf_log,
+      tf_base,
+      cli_debug,
+      latency_debug,
+      latency_limit,
+      watchdog_warning,
+      watchdog_timeout,
+      gr_wait,
+      hostname,
+      shutdown
+      );
+#undef COPY
+
+  ng->load_time = current_time();
+
+  if (new->router_id)
+    ng->router_id = new->router_id;
+  else if (old)
+  {
+    /* The startup router ID must be determined after start of device protocol,
+     * thus if old == NULL then we do nothing */
+    ng->router_id = og->router_id;
+
+    if (new->router_id_from)
     {
-      new->hostname = get_hostname(new->mem);
-
-      if (!new->hostname)
-        log(L_WARN "Cannot determine hostname");
+      u32 id = if_choose_router_id(new->router_id_from, og->router_id);
+      if (!id)
+	log(L_WARN "Cannot determine router ID, using old one");
+      else
+	ng->router_id = id;
     }
+  }
 
-  if (!old)
-    return 0;
+  atomic_store_explicit(&global_runtime, ng, memory_order_release);
 
-  if (!new->router_id)
-    {
-      new->router_id = old->router_id;
-
-      if (new->router_id_from)
-	{
-	  u32 id = if_choose_router_id(new->router_id_from, old->router_id);
-	  if (!id)
-	    log(L_WARN "Cannot determine router ID, using old one");
-	  else
-	    new->router_id = id;
-	}
-    }
-
-  return 0;
+  /* We have to wait until every reader surely doesn't read the old values */
+  synchronize_rcu();
 }
 
 static int
-config_do_commit(struct config *c, int type)
+config_do_commit(config_ref *cr, int type)
 {
   if (type == RECONFIG_UNDO)
     {
-      c = old_config;
+      OBSREF_SET(*cr, old_config);
       type = old_cftype;
     }
   else
     config_free(old_config);
 
-  old_config = config;
+  old_config = NULL;
   old_cftype = type;
-  config = c;
+
+  struct config *c = OBSREF_GET(*cr);
+  old_config = OBSREF_GET(config);
+
+  OBSREF_CLEAR(config);
+  OBSREF_SET(config, OBSREF_GET(*cr));
+
+  if (!c->hostname)
+    {
+      c->hostname = get_hostname(c->mem);
+
+      if (!c->hostname)
+        log(L_WARN "Cannot determine hostname");
+    }
 
   configuring = 1;
-  if (old_config && !config->shutdown)
+  if (old_config && !c->shutdown)
     log(L_INFO "Reconfiguring");
 
-  if (old_config)
-    old_config->obstacle_count++;
-
   DBG("filter_commit\n");
   filter_commit(c, old_config);
   DBG("sysdep_commit\n");
-  int force_restart = sysdep_commit(c, old_config);
+  sysdep_commit(c, old_config);
   DBG("global_commit\n");
-  force_restart |= global_commit(c, old_config);
+  global_commit(c, old_config);
+  mpls_commit(c, old_config);
   DBG("rt_commit\n");
   rt_commit(c, old_config);
   DBG("protos_commit\n");
-  protos_commit(c, old_config, force_restart, type);
+  protos_commit(c, old_config, type);
 
-  int obs = 0;
-  if (old_config)
-    obs = --old_config->obstacle_count;
+  /* Can be cleared directly? */
+  if (!old_config)
+    return 1;
 
-  DBG("do_commit finished with %d obstacles remaining\n", obs);
-  return !obs;
+  if (!callback_is_active(&old_config->obstacles_cleared))
+    return 0;
+
+  callback_cancel(&old_config->obstacles_cleared);
+  return 1;
 }
 
 static void
-config_done(void *unused UNUSED)
+config_done(void)
 {
-  if (config->shutdown)
+  struct config *c = OBSREF_GET(config);
+  ASSERT_DIE(c);
+
+  if (c->shutdown)
     sysdep_shutdown_done();
 
   configuring = 0;
+
   if (old_config)
     log(L_INFO "Reconfigured");
 
   if (future_cftype)
     {
       int type = future_cftype;
-      struct config *conf = future_config;
+      struct config *fc = OBSREF_GET(future_config);
+
+      if (type == RECONFIG_UNDO)
+      {
+	ASSERT_DIE(!fc);
+	ASSERT_DIE(old_config);
+	fc = old_config;
+      }
+
+      CONFIG_REF_LOCAL(conf, fc);
+
       future_cftype = RECONFIG_NONE;
-      future_config = NULL;
+      OBSREF_CLEAR(future_config);
 
       log(L_INFO "Reconfiguring to queued configuration");
-      if (config_do_commit(conf, type))
-	config_done(NULL);
+      if (config_do_commit(&conf, type))
+	config_done();
     }
 }
 
@@ -365,11 +410,11 @@ config_done(void *unused UNUSED)
  * are accepted.
  */
 int
-config_commit(struct config *c, int type, uint timeout)
+config_commit(config_ref *cr, int type, uint timeout)
 {
   if (shutting_down)
     {
-      config_free(c);
+      OBSREF_CLEAR(*cr);
       return CONF_SHUTDOWN;
     }
 
@@ -384,19 +429,19 @@ config_commit(struct config *c, int type, uint timeout)
       if (future_cftype)
 	{
 	  log(L_INFO "Queueing new configuration, ignoring the one already queued");
-	  config_free(future_config);
+	  OBSREF_CLEAR(future_config);
 	}
       else
 	log(L_INFO "Queueing new configuration");
 
       future_cftype = type;
-      future_config = c;
+      OBSREF_SET(future_config, OBSREF_GET(*cr));
       return CONF_QUEUED;
     }
 
-  if (config_do_commit(c, type))
+  if (config_do_commit(cr, type))
     {
-      config_done(NULL);
+      config_done();
       return CONF_DONE;
     }
   return CONF_PROGRESS;
@@ -460,8 +505,7 @@ config_undo(void)
     {
       if (future_cftype)
 	{
-	  config_free(future_config);
-	  future_config = NULL;
+	  OBSREF_CLEAR(future_config);
 
 	  log(L_INFO "Removing queued configuration");
 	  future_cftype = RECONFIG_NONE;
@@ -475,9 +519,11 @@ config_undo(void)
 	}
     }
 
-  if (config_do_commit(NULL, RECONFIG_UNDO))
+  CONFIG_REF_LOCAL_EMPTY(undo_conf);
+  if (config_do_commit(&undo_conf, RECONFIG_UNDO))
     {
-      config_done(NULL);
+      OBSREF_CLEAR(undo_conf);
+      config_done();
       return CONF_DONE;
     }
   return CONF_PROGRESS;
@@ -517,10 +563,7 @@ config_timeout(timer *t UNUSED)
 void
 config_init(void)
 {
-  config_pool = rp_new(&root_pool, "Configurations");
-
-  config_event = ev_new(config_pool);
-  config_event->hook = config_done;
+  config_pool = rp_new(&root_pool, the_bird_domain.the_bird, "Configurations");
 
   config_timer = tm_new(config_pool);
   config_timer->hook = config_timeout;
@@ -535,8 +578,6 @@ config_init(void)
 void
 order_shutdown(int gr)
 {
-  struct config *c;
-
   if (shutting_down)
     return;
 
@@ -545,17 +586,20 @@ order_shutdown(int gr)
   else
     log(L_INFO "Shutting down for graceful restart");
 
-  c = lp_alloc(config->mem, sizeof(struct config));
-  memcpy(c, config, sizeof(struct config));
+  struct config *c = lp_alloc(OBSREF_GET(config)->mem, sizeof(struct config));
+  memcpy(c, OBSREF_GET(config), sizeof(struct config));
   init_list(&c->protos);
   init_list(&c->tables);
+  init_list(&c->mpls_domains);
   init_list(&c->symbols);
+  obstacle_target_init(&c->obstacles, &c->obstacles_cleared, c->pool, "Config");
+  c->cli = (struct cli_config_list) {};
   memset(c->def_tables, 0, sizeof(c->def_tables));
-  HASH_INIT(c->sym_hash, c->pool, 4);
   c->shutdown = 1;
   c->gr_down = gr;
 
-  config_commit(c, RECONFIG_HARD, 0);
+  CONFIG_REF_LOCAL(cr, c);
+  config_commit(&cr, RECONFIG_HARD, 0);
   shutting_down = 1;
 }
 
