@@ -306,8 +306,8 @@ radv_prepare_prefix(struct radv_iface *ifa, struct radv_prefix *px,
   op->type = OPT_PREFIX;
   op->length = 4;
   op->pxlen = px->prefix.pxlen;
-  op->flags = (px->onlink ? OPT_PX_ONLINK : 0) |
-	      (px->autonomous ? OPT_PX_AUTONOMOUS : 0);
+  op->flags = (pc->onlink ? OPT_PX_ONLINK : 0) |
+    (pc->autonomous ? OPT_PX_AUTONOMOUS : 0);
   op->valid_lifetime = (ifa->ra->active || !pc->valid_lifetime_sensitive) ?
     htonl(pc->valid_lifetime) : 0;
   op->preferred_lifetime = (ifa->ra->active || !pc->preferred_lifetime_sensitive) ?
@@ -536,7 +536,7 @@ radv_sk_open(struct radv_iface *ifa)
   sk->data = ifa;
   sk->flags = SKF_LADDR_RX;
 
-  if (sk_open(sk) < 0)
+  if (sk_open(sk, ifa->ra->p.loop) < 0)
     goto err;
 
   /* We want listen just to ICMPv6 messages of type RS and RA */
@@ -554,7 +554,7 @@ radv_sk_open(struct radv_iface *ifa)
 
  err:
   sk_log_error(sk, ifa->ra->p.name);
-  rfree(sk);
+  sk_close(sk);
   return 0;
 }
 
